//==================================================================
//  DerefLnk.cpp
//    Bret S. Pehrson for Windows Developer's Journal
//==================================================================

#if defined(__BORLANDC__)
#   define  _alloca alloca
#else
/* disable unnecessary level 4 warnings */
#   pragma warning (disable : 4201)  //nameless struct/union
#   pragma warning (disable : 4514)  //unreferenced inline function
#   pragma warning (disable : 4214)  //bit field types other than int
#   pragma comment(lib,"kernel32.lib")
#   pragma comment(lib,"ole32.lib")
#   pragma comment(lib,"uuid.lib")
#   pragma comment(lib,"shell32.lib")
#endif

/* include files */
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <tchar.h>
#include <malloc.h>
#include <shlobj.h>
#include "dereflnk.h"

//------------------------------------------------------------------
//DerefShellLink
//  Parameters:
//    lpszLinkName - link name to dereference
//    lpszDest - destination to store dereferenced name
//    cchDest - size of destination buffer, in characters
//  Return: 
//    Number of characters (including NULL) for complete 
//    dereferenced name
//  Notes: 
//    lpszLinkName and lpszDest can point to the same buffer.
//    If NULL is specified for lpszDest, the dereferenced name
//    is not copied to the destination, but the return value is
//    still computed.
//------------------------------------------------------------------
UINT DerefShellLink(
    LPCTSTR lpszLinkName, 
    LPTSTR lpszDest, 
    UINT cchDest)
{
    /* STACK variables */
    const LPCTSTR   lpszExt = _tcsrchr(lpszLinkName, '.');
    WIN32_FIND_DATA fd;
    IPersistFile *  pPersistFile = NULL; 
    IShellLink *    pShellLink = NULL; 
    LPCOLESTR       lpOleStrLinkName;
    HRESULT         hResult;
    TCHAR           szPath[MAX_PATH+1];
    UINT            cchDeref = 0;
    UINT            cchString;
    #ifndef UNICODE
    UINT            cbAlloc;
    #endif //!UNICODE

    /* initialization */
    fd.cFileName[0] = NULL;
    szPath[0]       = NULL;
    /* dereference as necessary */
    if ((lpszExt!=NULL) && (lstrcmpi(lpszExt, TEXT(".lnk"))==0))
    {
        /* initialize COM */
        hResult = CoInitialize(NULL);
        if (SUCCEEDED(hResult))
        {
            /* create an instance of shell link interface */
            hResult = CoCreateInstance(
                CLSID_ShellLink, 
                NULL, 
                CLSCTX_INPROC_SERVER, 
                IID_IShellLink, 
                (LPVOID *)&pShellLink);
            if (SUCCEEDED(hResult))
            {
                /* get the persist file interface */
                hResult = pShellLink->QueryInterface(
                    IID_IPersistFile, 
                    (LPVOID *)&pPersistFile);
                if (SUCCEEDED(hResult))
                {
                    /* convert the link name to unicode 
                     * as necessary */
                    #ifdef UNICODE
                    lpOleStrLinkName = lpszLinkName;
                    #else //!UNICODE
                    cchString = lstrlen(lpszLinkName) + 1;
                    cbAlloc = cchString * sizeof(OLECHAR);
                    lpOleStrLinkName = (LPOLESTR)_alloca(cbAlloc);
                    MultiByteToWideChar(
                        CP_ACP, 
                        MB_PRECOMPOSED, 
                        lpszLinkName, 
                        cchString, 
                        (LPOLESTR)lpOleStrLinkName, 
                        cchString);
                    #endif //!UNICODE
                    /* load the persist file with the link */
                    hResult = pPersistFile->Load(
                        lpOleStrLinkName, 
                        0);
                    if (SUCCEEDED(hResult))
                    {
                        /* resolve the link */
                        hResult = pShellLink->Resolve(
                            NULL, 
                            SLR_NO_UI);
                        if (SUCCEEDED(hResult))
                        {
                            /* get the path from the link */
                            hResult = pShellLink->GetPath(
                                szPath, 
                                sizeof(szPath)/sizeof(szPath[0]), 
                                &fd, 
                                0);
                            if (hResult == S_OK)
                            {
                                /* copy to destination */
                                cchDeref = lstrlen(szPath) + 1;
                                if (lpszDest != NULL)
                                {
                                    lstrcpyn(
                                        lpszDest,
                                        szPath,
                                        cchDest);
                                }
                            }
                        }
                    }
                    pPersistFile->Release(); pPersistFile = NULL;
                }
                pShellLink->Release(); pShellLink = NULL;
            }
            CoUninitialize();
        }
    }
    else
    {
        /* not a shell link, simply copy source to destination */
        if ((lpszLinkName!=lpszDest) && (lpszDest!= NULL))
        {
            lstrcpyn(lpszDest, lpszLinkName, cchDest);
        }
        cchDeref = lstrlen(lpszLinkName) + 1;
    }
    /* done */
    return cchDeref;
}
